﻿using System;
using System.Collections.Generic;
using System.Xml.Serialization;

namespace AOUtils.Libraries
{
    public class OptimizedItem : IComparable, IEquatable<OptimizedItem>
    {
        public OptimizedItem()
        {
            OnWear = new OnWear();
            AttackSkills = new AttackValueCollection();
            ToUse = new OptimizedRequirementCollection();
            ToWear = new OptimizedRequirementCollection();
            ToWield = new OptimizedRequirementCollection();
        }
        public static explicit operator OptimizedItem(Item item)
        {
            OptimizedItem oitem = new OptimizedItem();
            oitem.Id = item.Id;
            oitem.Name = item.Name;
            oitem.Ql = item.Ql;
            oitem.Type = item.Type;
            oitem.Unique = (item.Flags & Flags.Unique) > 0;
            oitem.Icon = item.Icon;
            oitem.Flags = item.Flags;
            oitem.Slot = item.Slot;
            oitem.EquipmentPage = item.EquipmentPage;
            oitem.MultiRanged = item.MultiRanged;
            oitem.MultiMelee = item.MultiMelee;

            foreach (AttackValue val in item.AttackValues)
            {
                oitem.AttackSkills.Add(val);
                /*if ((item.EquipmentPage & EquipmentPage.Weapon) > 0 && (val.Name & (AttackValueTypes.Ranged | AttackValueTypes.Melee)) == 0)
                    Console.WriteLine(item.Name);*/
            }

            if (item.Events.Contains(EventType.OnWear))
            {
                foreach (Function func in item.Events[EventType.OnWear].Functions)
                {
                    TargetType ttype = (TargetType)Enum.Parse(typeof(TargetType), func.Target);
                    if (!oitem.OnWear.Functions.Contains(func.Name))
                    {
                        FunctionList flist = new FunctionList();
                        flist.Type = func.Name;
                        oitem.OnWear.Functions.Add(flist);
                    }
                    Target t = new Target();
                    t.Type = (TargetType)Enum.Parse(typeof(TargetType), func.Target);
                    t.Arguments = new List<Argument>();
                    t.Arguments.AddRange(func.Arguments);
                    if (!oitem.OnWear.Functions[func.Name].Targets.Contains(ttype))
                    {
                        TargetList tlist = new TargetList();
                        tlist.Type = ttype;
                        oitem.OnWear.Functions[func.Name].Targets.Add(tlist);
                    }
                    foreach (Argument arg in func.Arguments)
                    {
                        List<OptimizedRequirement> reqs = new List<OptimizedRequirement>();
                        foreach (Requirement req in func.Requirements)
                            reqs.Add(req);
                        oitem.OnWear.Functions[func.Name].Targets[ttype].Arguments.Add(new OptimizedArgument(arg, reqs));
                    }
                }
            }
            foreach (Action action in item.Actions)
            {
                if (item.Actions.Contains(ActionType.ToUse))
                {
                    List<OptimizedRequirement> reqs = new List<OptimizedRequirement>();
                    foreach (Requirement req in action.Requirements)
                        oitem.ToUse.Requirements.Add(req);
                }
                if (item.Actions.Contains(ActionType.ToWear))
                {
                    List<OptimizedRequirement> reqs = new List<OptimizedRequirement>();
                    foreach (Requirement req in action.Requirements)
                        oitem.ToWear.Requirements.Add(req);
                }
                if (item.Actions.Contains(ActionType.ToWield))
                {
                    List<OptimizedRequirement> reqs = new List<OptimizedRequirement>();
                    foreach (Requirement req in action.Requirements)
                        oitem.ToWield.Requirements.Add(req);
                }
            }
            return oitem;
        }

        [XmlAttribute("id")]
        public int Id
        {
            get;
            set;
        }
        [XmlAttribute("name")]
        public string Name
        {
            get;
            set;
        }
        [XmlAttribute("ql")]
        public int Ql
        {
            get;
            set;
        }
        [XmlAttribute("itemtype")]
        public ItemType Type
        {
            get;
            set;
        }
        [XmlAttribute("unique")]
        public bool Unique
        {
            get;
            set;
        }
        [XmlElement("icon")]
        public int Icon
        {
            get;
            set;
        }
        [XmlElement("flags")]
        public Flags Flags
        {
            get;
            set;
        }
        [XmlElement("multiranged")]
        public int MultiRanged
        {
            get;
            set;
        }
        [XmlElement("multimelee")]
        public int MultiMelee
        {
            get;
            set;
        }
        [XmlArray("attackskills")]
        [XmlArrayItem("attackskill",typeof(AttackValue))]
        public AttackValueCollection AttackSkills
        {
            get;
            set;
        }
        [XmlIgnore]
        public bool IsDualWieldable
        {
            get
            {
                return MultiMelee > 0 || MultiRanged > 0;
            }
        }
        [XmlIgnore]
        public AttackValueTypes WeaponType
        {
            get
            {
                if (AttackSkills.Count == 0)
                    return AttackValueTypes.None;
                AttackValue value = new AttackValue();
                foreach (AttackValue val in AttackSkills)
                {
                    if (val.Percent > value.Percent)
                        value = val;
                }
                if ((value.Name & AttackValueTypes.Ranged) > 0)
                    return AttackValueTypes.Ranged;
                else if ((value.Name & AttackValueTypes.Melee) > 0)
                    return AttackValueTypes.Melee;
                return AttackValueTypes.Other;
            }
        }
        [XmlIgnore]
        public bool Social
        {
            get
            {
                return (Flags & Flags.ItemSocialArmour) > 0;
            }
        }
        [XmlIgnore]
        public bool NoDrop
        {
            get
            {
                return (Flags & Flags.NoDrop) > 0;
            }
        }
        [XmlElement("slot")]
        public Slot Slot
        {
            get;
            set;
        }
        [XmlElement("equipmentpage")]
        public EquipmentPage EquipmentPage
        {
            get;
            set;
        }
        [XmlElement("onwear")]
        public OnWear OnWear
        {
            get;
            set;
        }
        [XmlElement("touse")]
        public OptimizedRequirementCollection ToUse
        {
            get;
            set;
        }
        [XmlElement("towear")]
        public OptimizedRequirementCollection ToWear
        {
            get;
            set;
        }
        [XmlElement("towield")]
        public OptimizedRequirementCollection ToWield
        {
            get;
            set;
        }
        [XmlIgnore]
        public string SortSkill
        {
            get;
            set;
        }

        public SortedDictionary<string, int> GetAbsoluteSkillValues()
        {
            SortedDictionary<string, int> skills = GetSkillValues();
            SortedDictionary<string, int> modskills = new SortedDictionary<string, int>();
            foreach (KeyValuePair<string, int> skill in skills)
            {
                switch (skill.Key)
                {
                    case "NPCostModifier":
                    case "SkillLockModifier":
                        modskills.Add(skill.Key, skill.Value * -1);
                        break;
                    default:
                        modskills.Add(skill.Key, skill.Value);
                        break;
                }
            }
            return modskills;
        }

        public SortedDictionary<string, int> GetSkillValues()
        {
            SortedDictionary<string, int> skills = new SortedDictionary<string, int>();
            if (OnWear.Functions.Contains(FunctionType.Modify) && OnWear.Functions[FunctionType.Modify].Targets.Contains(TargetType.Wearer))
            {
                KeyValuePair<string, int> kvp;
                for (int x = OnWear.Functions[FunctionType.Modify].Targets[TargetType.Wearer].Arguments.Count - 1; x >= 0; x -= 2)
                {
                    int i;
                    if (int.TryParse(OnWear.Functions[FunctionType.Modify].Targets[TargetType.Wearer].Arguments[x].Value, out i))
                    {
                        kvp = new KeyValuePair<string, int>(
                            OnWear.Functions[FunctionType.Modify].Targets[TargetType.Wearer].Arguments[x - 1].Value,
                            int.Parse(OnWear.Functions[FunctionType.Modify].Targets[TargetType.Wearer].Arguments[x].Value));
                    }
                    else
                    {
                        kvp = new KeyValuePair<string, int>(
                            OnWear.Functions[FunctionType.Modify].Targets[TargetType.Wearer].Arguments[x].Value,
                            int.Parse(OnWear.Functions[FunctionType.Modify].Targets[TargetType.Wearer].Arguments[x - 1].Value));
                    }
                    if (!skills.ContainsKey(kvp.Key))
                        skills.Add(kvp.Key, kvp.Value);
                }
            }
            if (OnWear.Functions.Contains(FunctionType.Modify) && OnWear.Functions[FunctionType.Modify].Targets.Contains(TargetType.Target))
            {
                KeyValuePair<string, int> kvp;
                for (int x = OnWear.Functions[FunctionType.Modify].Targets[TargetType.Target].Arguments.Count - 1; x >= 0; x -= 2)
                {
                    int i;
                    if (int.TryParse(OnWear.Functions[FunctionType.Modify].Targets[TargetType.Target].Arguments[x].Value, out i))
                    {
                        kvp = new KeyValuePair<string, int>(
                            OnWear.Functions[FunctionType.Modify].Targets[TargetType.Target].Arguments[x - 1].Value,
                            int.Parse(OnWear.Functions[FunctionType.Modify].Targets[TargetType.Target].Arguments[x].Value));
                    }
                    else
                    {
                        kvp = new KeyValuePair<string, int>(
                            OnWear.Functions[FunctionType.Modify].Targets[TargetType.Target].Arguments[x].Value,
                            int.Parse(OnWear.Functions[FunctionType.Modify].Targets[TargetType.Target].Arguments[x - 1].Value));
                    }
                    if (!skills.ContainsKey(kvp.Key))
                        skills.Add(kvp.Key, kvp.Value);
                }
            }
            return skills;
        }

        public List<OptimizedRequirement> GetRequirements(string statistic)
        {
            List<OptimizedRequirement> reqs = new List<OptimizedRequirement>();
            foreach (OptimizedRequirement req in ToUse.Requirements)
                if (req.Statistic.Equals(statistic))
                    reqs.Add(req);
            foreach (OptimizedRequirement req in ToWear.Requirements)
                if (req.Statistic.Equals(statistic))
                    reqs.Add(req);
            foreach (OptimizedRequirement req in ToWield.Requirements)
                if (req.Statistic.Equals(statistic))
                    reqs.Add(req);
            return reqs;
        }

        [XmlIgnore]
        public List<OptimizedRequirement> HasFactionReq
        {
            get
            {
                return GetRequirements("Side");
            }
        }
        [XmlIgnore]
        public List<OptimizedRequirement> HasBreedReq
        {
            get
            {
                return GetRequirements("Breed");
            }
        }
        [XmlIgnore]
        public List<OptimizedRequirement> HasProfessionReq
        {
            get
            {

                List<OptimizedRequirement> list = GetRequirements("Profession");
                list.AddRange(GetRequirements("VisualProfession"));
                if (list.Count == 0 && (this.Type & ItemType.Implant) > 0)
                {
                    OptimizedRequirement req = new OptimizedRequirement();
                    req.Operator = OperatorType.NotEqualTo;
                    req.Value = "Shade";
                    req.Statistic = "Profession";
                    list.Add(req);
                }
                else if (list.Count == 0 && (this.Type & ItemType.Spirit) > 0)
                {
                    OptimizedRequirement req = new OptimizedRequirement();
                    req.Operator = OperatorType.EqualTo;
                    req.Value = "Shade";
                    req.Statistic = "Profession";
                    list.Add(req);
                }

                return list;
            }
        }

        #region IComparable Members

        public int CompareTo(object obj)
        {
            if (obj is OptimizedItem)
            {
                OptimizedItem item = obj as OptimizedItem;
                if (SortSkill != null && !SortSkill.Equals(""))
                {
                    return GetAbsoluteSkillValues()[SortSkill].CompareTo(item.GetAbsoluteSkillValues()[SortSkill]);
                }
                return item.Name.CompareTo(Name);
            }
            return -1;
        }

        #endregion

        #region IEquatable<OptimizedItem> Members

        public bool Equals(OptimizedItem other)
        {
            return this.Name.Equals(other.Name);
        }

        public override bool Equals(object obj)
        {
            if (obj is OptimizedItem)
                return this.Name.Equals(((OptimizedItem)obj).Name);
            return false;
        }

        public override int GetHashCode()
        {
            return this.Name.GetHashCode();
        }
        #endregion

        public static bool operator >(OptimizedItem x, OptimizedItem y)
        {
            if (x == null)
                return false;
            if (x != null && y == null)
                return true;
            return x.GetAbsoluteSkillValues()[x.SortSkill] > y.GetAbsoluteSkillValues()[y.SortSkill];
        }

        public static bool operator <(OptimizedItem x, OptimizedItem y)
        {
            if (x == null)
                return false;
            if (x != null && y == null)
                return true;
            return x.GetAbsoluteSkillValues()[x.SortSkill] < y.GetAbsoluteSkillValues()[y.SortSkill];
        }

        public static bool operator >=(OptimizedItem x, OptimizedItem y)
        {
            return !(x.GetAbsoluteSkillValues()[x.SortSkill] < y.GetAbsoluteSkillValues()[y.SortSkill]);
        }

        public static bool operator <=(OptimizedItem x, OptimizedItem y)
        {
            return !(x.GetAbsoluteSkillValues()[x.SortSkill] > y.GetAbsoluteSkillValues()[y.SortSkill]);
        }

        public static bool operator ==(OptimizedItem x, OptimizedItem y)
        {
            if (Object.ReferenceEquals(x, y))
                return true;
            if (((object)x == null) || ((object)y == null))
                return false;

            return x.GetAbsoluteSkillValues()[x.SortSkill] == y.GetAbsoluteSkillValues()[y.SortSkill];
        }

        public static bool operator !=(OptimizedItem x, OptimizedItem y)
        {
            return !(x == y);
        }
    }
}
